require "iuplua"
require "iupluacontrols"
require "midiwire"

-- now, this can actually play Blue monday!
-- see https://www.youtube.com/watch?v=Y_hk5P1X7sc&t=164s

LEN = 33
step = 1
mout = midiwire.open_out(0)
active_pat = 1

instruments = {
	[1] = {"clap", 39},
	[2] = {"open hat", 46},
	[3] = {"snare", 40},
	[4] = {"bass", 36},
}

patterns = {
	{},
	{},
	{},
	{}
}

-- GUI definition

pat_a = iup.button{title="PAT A", fontstyle="Bold"}
pat_b = iup.button{title="PAT B"}
pat_c = iup.button{title="PAT C"}
pat_d = iup.button{title="PAT D"}
pat_rec = iup.button{title="REC"}
pat_load = iup.button{title="load"}
pat_save = iup.button{title="save"}

matrix = iup.matrix{numcol=32, numlin=9, widthdef=10}
matrix.resizematrix = "YES"

matrix:setcell(0,0,"DRUMsheet")
for i=1,32 do
	matrix:setcell(0,i,i)
end
for i=1,#instruments do
	matrix:setcell(i,0,instruments[i][1])
end

playbtn = iup.button{title="play"}
stopbtn = iup.button{title="stop"}

-- TODO - maskovat tyhle věci a validovat ať to necrashuje

bpm = iup.text{spin="YES", value="120", spinmin=30, spinmax=240} -- todo: min max sensible
factor = iup.text{spin="YES", value="4", spinmin=1, spinmax=8}
len = iup.text{spin="YES", value="32", spinmin=8, spinmax=32}

dlg = iup.dialog{
	iup.vbox{
		iup.hbox{pat_a, pat_b, pat_c, pat_d, pat_rec ,pat_load, pat_save; gap=5},
		iup.hbox{matrix},
		iup.hbox{playbtn, stopbtn, iup.label{title="tempo: "} ,  bpm, iup.label{title="/"}, factor, iup.label{title=" || len: "}, len; alignment="ACENTER", gap="5"}
	}
	; title = "DRUMsheet", size = "560x170"
}

-- GUI actions

timer = iup.timer{time=500}
timer.run="NO"

function timer:action_cb()
	-- ONE STEP of sequencer
	if step==1 then
		matrix:setcell(9, LEN-1, "")
	else
		matrix:setcell(9, step-1, "") -- clean from previous step
	end
	matrix:setcell(9, step, "^")

	for i=1,#instruments do
		if matrix:getcell(i, step)=="x" then
			midiwire.drum(mout, instruments[i][2])
		end
	end
	
	step = ((step + 1) % LEN)
	if step==0 then 
		step=1 
	end
	matrix.redraw="YES"
end

function playbtn:action()
	bpm.active="NO"
	factor.active="NO"
	len.active="NO"
	timer.run="YES"
end

function stopbtn:action()
	bpm.active="YES"
	factor.active="YES"
	len.active="YES"
	timer.run="NO"
	-- todo  - tohle lépe
	if step==1 then
		matrix:setcell(9, LEN-1, "")
	else
		matrix:setcell(9, step-1, "") -- clean from previous step
	end
	matrix.redraw="YES"
	step = 1
end

function adjust_time(bpm, factor)
	timer.time = math.floor((1000*60) / (bpm*factor) )
end

function bpm:valuechanged_cb()
	if tonumber(bpm.value) and tonumber(factor.value) then
		adjust_time(tonumber(bpm.value), tonumber(factor.value))
	end
end

function factor:valuechanged_cb()
	if tonumber(bpm.value) and tonumber(factor.value) then
		adjust_time(tonumber(bpm.value), tonumber(factor.value))
	end
end

function len:valuechanged_cb()
	if tonumber(len.value) then
		LEN = tonumber(len.value) + 1
		for i=1,LEN do
			matrix:setcell(0, i, i)
		end
		if LEN<33 then
			for i=LEN,32 do
				matrix:setcell(0, i, "")
			end
		end
		matrix.redraw="YES"
	end
end


-- easier editing of patterns

function matrix:action_cb(key, x, y, edition, value)
	if key==iup.K_SP then
		matrix:setcell(x, y, "")
		matrix.redraw="YES"
		return iup.IGNORE
	elseif key==iup.K_X or key==iup.K_x then
		matrix:setcell(x, y, "x")
		matrix.redraw="YES"
		return iup.IGNORE
	end
end


-- patterns changing, saving and loading

function pat_rec:action()
	patterns[active_pat] = {}
	for x=1,8 do
		patterns[active_pat][x] = {}
		for y=1,32 do
			patterns[active_pat][x][y] = matrix:getcell(x,y) or ""
		end
		-- print(table.concat(patterns[active_pat][x], ","))
	end
end

function pat_recall(patno)
	local was_running = timer.run == "YES"
	timer.run="NO" -- pause while recalling to not glitch
	for x=1,8 do
		patterns[patno][x] = patterns[patno][x] or {}
		for y=1,32 do
			matrix:setcell(x, y, patterns[patno][x][y] or "")
		end
		-- print(table.concat(patterns[active_pat][x], ","))
	end
	active_pat = patno
	if was_running then
		step = 1
		timer.run="YES"
	end
	matrix.redraw="YES"
end

function pat_a:action()
	pat_a.fontstyle="Bold"; pat_b.fontstyle=""; pat_c.fontstyle=""; pat_d.fontstyle=""
	pat_recall(1)
end

function pat_b:action()
	pat_a.fontstyle=""; pat_b.fontstyle="Bold"; pat_c.fontstyle=""; pat_d.fontstyle=""
	pat_recall(2)
end

function pat_c:action()
	pat_a.fontstyle=""; pat_b.fontstyle=""; pat_c.fontstyle="Bold"; pat_d.fontstyle=""
	pat_recall(3)
end


function pat_d:action()
	pat_a.fontstyle=""; pat_b.fontstyle=""; pat_c.fontstyle=""; pat_d.fontstyle="Bold"
	pat_recall(4)
end

-- saving & loading

function _pat_write(fh, patno)
	fh:write("PAT,"..patno.."\n")
	for i=1,#patterns[patno] do
		fh:write(table.concat(patterns[patno][i],",") .. "\n")
	end
end

function pat_save:action()
	local fname, status = iup.GetFile("./*.csv")
	if status > -1 then
		local fh = io.open(fname, "w+")
		fh:write("DRUMSHEET\n")
		fh:write("tempo,"..bpm.value.."\n")
		fh:write("factor,"..factor.value.."\n")
		fh:write("len,"..len.value.."\n")
		fh:write("INSTRUMENTS\n")
		for i=1,#instruments do
			fh:write(instruments[i][1] .. "," .. instruments[i][2] .. "\n")
		end
		-- todo pat_write nebo loop
		if #patterns[1]>1 then _pat_write(fh, 1) end
		if #patterns[2]>1 then _pat_write(fh, 2) end
		if #patterns[3]>1 then _pat_write(fh, 3) end
		if #patterns[4]>1 then _pat_write(fh, 4) end
		fh:close()
	end
end


-- main program

adjust_time(120, 4)
dlg:showxy( iup.CENTER, iup.CENTER )

if (iup.MainLoopLevel()==0) then
  iup.MainLoop()
end